from typing import Dict
from scrapy.item import Item

class ItemValidator(object):

    ITEM_COVERAGE_ENABLED = True
    INVALID_ITEM_URLS_LOGGING_ENABLED = False
    MAX_ITEM_URLS = 1000

    def __init__(self):
        self.item_coverage = {
            '_SOP_OVERAL_STATS': {
                    'num_items': 0,
                    'num_invalid_items': 0,
                    'num_total_fields': 0,
                    'num_invalid_fields': 0,
                }
        }
        self.items = 0
        self.invalid_items = 0
        self.invalid_items_urls = {}


    def extract_name_fields_item(item):
        return 

    def validate(self, request_response_object, item):
        if ItemValidator.ITEM_COVERAGE_ENABLED:
            self.increment_items()
            if isinstance(item, Item):
                item_name = ItemValidator.get_item_name(item)
                fields = ItemValidator.get_item_fields(item)
                field_keys = fields.keys()
            if isinstance(item, Dict):
                item_name = 'Unknown'
                field_keys = item.keys()
            if item_name is not None and field_keys is not None:
                domain = request_response_object.get_domain()
                invalid_fields = []
                valid_item = True
                self.check_item_exists(domain, item_name, field_keys)
                self.item_coverage[domain][item_name]['num_items'] += 1
                self.increment_total_fields(field_keys)
                for k in field_keys:
                    if(item.get(k) is not None and item.get(k) != ''):
                        self.item_coverage[domain][item_name]['coverage'][k] += 1
                    else:
                        valid_item = False
                        self.increment_invalid_fields()
                        invalid_fields.append(k)
                
                if valid_item is False:
                    self.item_coverage[domain][item_name]['num_invalid_items'] += 1
                    self.increment_invalid_items()
                if ItemValidator.INVALID_ITEM_URLS_LOGGING_ENABLED and len(invalid_fields) > 0:
                    self.log_invalid_item_url(request_response_object.get_real_url(), item_name, invalid_fields)


    def check_item_exists(self, domain, item_name, field_keys):
        if self.item_coverage.get(domain) is None:
            self.item_coverage[domain] = {}
        if self.item_coverage[domain].get(item_name) is None:
            self.item_coverage[domain][item_name] = {
                'coverage': {},
                'num_fields': 0,
                'num_items': 0,
                'num_invalid_items': 0,
            }
            self.item_coverage[domain][item_name]['num_fields'] = len(field_keys)
            for k in field_keys:
                self.item_coverage[domain][item_name]['coverage'][k] = 0
                

    def log_invalid_item_url(self, url, item_name, invalid_fields):
        if self.invalid_items_urls.get(item_name) is None:
            self.invalid_items_urls[item_name] = {}
        missing_fields_string = ItemValidator.generate_fields_key(invalid_fields)
        if self.invalid_items_urls[item_name].get(missing_fields_string) is None:
            self.invalid_items_urls[item_name][missing_fields_string] = []
        if url not in self.invalid_items_urls[item_name][missing_fields_string] and len(self.invalid_items_urls[item_name][missing_fields_string]) < ItemValidator.MAX_ITEM_URLS:
            self.invalid_items_urls[item_name][missing_fields_string].append(url)


    def increment_total_fields(self, fields):
        self.item_coverage['_SOP_OVERAL_STATS']['num_total_fields'] += len(fields)

    def increment_invalid_fields(self):
        self.item_coverage['_SOP_OVERAL_STATS']['num_invalid_fields'] += 1

    def increment_items(self):
        self.items += 1
        self.item_coverage['_SOP_OVERAL_STATS']['num_items'] += 1

    def increment_invalid_items(self):
        self.invalid_items += 1
        self.item_coverage['_SOP_OVERAL_STATS']['num_invalid_items'] += 1

    def get_item_coverage_data(self):
        return self.item_coverage

    def get_num_items(self):
        return self.items

    def get_num_invalid_items(self):
        return self.invalid_items

    def get_invalid_items_urls(self):
        return self.invalid_items_urls
    
    def get_field_coverage(self):
        overall_stats = self.item_coverage.get('_SOP_OVERAL_STATS')
        if overall_stats is None: return 0
        if overall_stats.get('num_total_fields', 0) == 0: return 0
        valid_fields = overall_stats.get('num_total_fields') - overall_stats.get('num_invalid_fields')
        return round((valid_fields / overall_stats.get('num_total_fields'))*100)
    

    @staticmethod
    def get_item_fields(item):
        return item.fields

    @staticmethod
    def get_item_name(item):
        return item.__class__.__name__

    @staticmethod
    def generate_fields_key(fields):
        missing_fields_string = ''
        for field in fields:
            if len(missing_fields_string) > 0:
                 missing_fields_string += '&&'
            missing_fields_string += field
        return missing_fields_string








        
        


        

        






from scrapeops_scrapy.core.api import SOPSRequest 


class ProxyPortStringNormalizer(object):

    def __init__(self):
        pass
    

    @staticmethod
    def run_proxy_string_normalization(request_response_object, normalization_actions):

        if normalization_actions is not None:
            for action_type, actions in normalization_actions.items():
                if actions is not None:
                    if action_type == 'username':
                        for action_block in actions:
                            updated = ProxyPortStringNormalizer.process_action(request_response_object.get_normalized_proxy_port_username(), action_block)
                            request_response_object.set_normalized_proxy_port_username(updated)

                    if action_type == 'password':
                        for action_block in actions:
                            updated = ProxyPortStringNormalizer.process_action(request_response_object.get_normalized_proxy_port_password(), action_block)
                            request_response_object.set_normalized_proxy_port_password(updated)

                    if action_type == 'host':
                        for action_block in actions:
                            updated = ProxyPortStringNormalizer.process_action(request_response_object.get_normalized_proxy_port_host(), action_block)
                            request_response_object.set_normalized_proxy_port_host(updated)

                    if action_type == 'port':
                        for action_block in actions:
                            updated = ProxyPortStringNormalizer.process_action(request_response_object.get_normalized_proxy_port_port(), action_block)
                            request_response_object.set_normalized_proxy_port_port(updated)

                    if action_type == 'headers':
                        for action_block in actions:
                            updated = ProxyPortStringNormalizer.process_action(request_response_object.get_proxy_port_headers(), action_block)
                            if updated is not None:
                                request_response_object.update_normalized_proxy_port_header_string(updated)


    @staticmethod
    def process_action(inputValue, action_block):

        if action_block.get('action') == 'contains_replace':
            return ProxyPortStringNormalizer.contains_replace(inputValue, action_block)

        if action_block.get('action') == 'contains_replace_all':
            return ProxyPortStringNormalizer.contains_replace_all(inputValue, action_block)

        if action_block.get('action') == 'not_contains_replace_all':
            return ProxyPortStringNormalizer.not_contains_replace_all(inputValue, action_block)
            
        if action_block.get('action') == 'replace_key_value':
            return ProxyPortStringNormalizer.replace_key_value(inputValue, action_block)

        if action_block.get('action') == 'replace_key_seperator_value':
            return ProxyPortStringNormalizer.replace_key_seperator_value(inputValue, action_block)

        if action_block.get('action') == 'check_headers_contains':
            return ProxyPortStringNormalizer.check_headers_contains(inputValue, action_block)

        if action_block.get('action') == 'not_ends_in_replace':
            return ProxyPortStringNormalizer.not_ends_in_replace(inputValue, action_block)
        
        if action_block.get('action') == 'ends_in_replace':
            return ProxyPortStringNormalizer.ends_in_replace(inputValue, action_block)

        if action_block.get('action') == 'equals_replace':
            return ProxyPortStringNormalizer.equals_replace(inputValue, action_block)

        if action_block.get('action') == 'not_equals_replace':
            return ProxyPortStringNormalizer.not_equals_replace(inputValue, action_block)

        if action_block.get('action') == 'is_none_replace':
            return ProxyPortStringNormalizer.is_none_replace(inputValue, action_block)

        if action_block.get('action') == 'in_list_replace':
            return ProxyPortStringNormalizer.in_list_replace(inputValue, action_block)

        if action_block.get('action') == 'not_in_list_replace':
            return ProxyPortStringNormalizer.not_in_list_replace(inputValue, action_block)
        
        
    """
        Conditional Checks
    """
    @staticmethod
    def conditional_checks(inputString, condition=None):
        if condition is not None and condition.get('type') is not None:
            
            ## If substring in string
            if condition.get('type') == "contains":
                if condition.get('value') in inputString:
                    return True
                return False

            if condition.get('type') == "not_contains":
                if condition.get('value') not in inputString:
                    return True
                return False

            if condition.get('type') == "equals":
                if condition.get('value') == inputString:
                    return True
                return False

            if condition.get('type') == "not_equal":
                if condition.get('value') != inputString:
                    return True
                return False

            if condition.get('type') == "not_none":
                if inputString is not None:
                    return True
                return False

            
            ## If all tests fail
            return False
        
        return True
    
    @staticmethod
    def get_condition_arguements(action_block):
        return action_block.get('condition'), action_block.get('arguements')


    """
        Actions
    """

    @staticmethod
    def replace_key_value(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        substring = arguements.get('substring')
        string_seperator = arguements.get('seperator')
        replacement = arguements.get('replacement')

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            outputString = inputString
            splitString = inputString.split(string_seperator)
            for el in splitString:
                if substring.startswith('**'):
                    if el.split('=')[0] == substring[2:]:
                        outputString = outputString.replace(el, replacement)
                elif substring in el:
                    outputString = outputString.replace(el, replacement)
            return outputString
        
        return inputString

    
    @staticmethod
    def replace_key_seperator_value(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        substring = arguements.get('substring')
        string_seperator = arguements.get('seperator')
        replacement = arguements.get('replacement')
        next_value = arguements.get('next_value')

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            outputString = inputString
            splitString = inputString.split(string_seperator)
            for i, el in enumerate(splitString):
                if substring == el:
                    if i + int(next_value) <= len(splitString):
                        outputString = outputString.replace(string_seperator + splitString[i + int(next_value)], replacement)
            return outputString
        
        return inputString


    @staticmethod
    def check_headers_contains(inputheaders, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)
        value = condition.get('value')
        if inputheaders.get(value) is not None:
            header_value = inputheaders.get(value)
            if type(header_value) is list:
                header_value = header_value[0]
            value_check = arguements.get('check_type')
            if value_check == 'equals':
                if header_value == arguements.get('value'):
                    return arguements.get('addition')
            if value_check == 'not_equal':
                if header_value != arguements.get('value'):
                    return arguements.get('addition')
            if value_check is None:
                return arguements.get('addition')
        return None

    @staticmethod
    def not_ends_in_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if inputString.endswith(substring) is False:
                return arguements.get('replacement')
        
        return inputString

    @staticmethod
    def ends_in_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if inputString.endswith(substring):
                return arguements.get('replacement')
        
        return inputString


    @staticmethod
    def not_equals_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if inputString != substring:
                return arguements.get('replacement')
        
        return inputString


    @staticmethod
    def equals_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if inputString == substring:
                return arguements.get('replacement')
        
        return inputString



    @staticmethod
    def contains_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            replacement = arguements.get('replacement')
            return inputString.replace(substring, replacement)
        
        return inputString

    
    @staticmethod
    def not_contains_replace_all(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if substring not in inputString:
                return arguements.get('replacement')
        
        return inputString

    @staticmethod
    def contains_replace_all(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if isinstance(inputString, str) is False:
            inputString = str(inputString)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            substring = arguements.get('substring')
            if substring in inputString:
                return arguements.get('replacement')
        
        return inputString

    
    @staticmethod
    def is_none_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            if inputString is None:
                return arguements.get('replacement')
        
        return inputString

    @staticmethod
    def in_list_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            list = arguements.get('list', '').split(',')
            if inputString in list:
                return arguements.get('replacement')
        
        return inputString

    
    @staticmethod
    def not_in_list_replace(inputString, action_block):
        condition, arguements = ProxyPortStringNormalizer.get_condition_arguements(action_block)

        if ProxyPortStringNormalizer.conditional_checks(inputString, condition=condition):
            list = arguements.get('list', '').split(',')
            if inputString not in list:
                return arguements.get('replacement')
        
        return inputString


    @staticmethod
    def proxy_port_test(proxy_dict, request_response_object, data, valid=False):
        if valid:
            proxy_name = request_response_object.get_proxy_port_name()
            test_request = data.get('test_request')
            proxy_port_details = data.get('proxy_port_details')
            proxy_setup_key = proxy_port_details.get('proxy_setup_key')
            if proxy_dict[proxy_name].get('sops_test_request') is None:
                proxy_dict[proxy_name]['sops_test_request'] = test_request
            test_request_count = proxy_dict[proxy_name]['sops_test_request'].get('count', 0)
            if test_request.get('send') and test_request_count < test_request.get('max_count', 1):
                proxy_dict[proxy_name]['sops_test_request']['count'] = test_request_count + 1
                json = SOPSRequest().proxy_test_request(test_request.get('url'), request_response_object)
                json['test_id'] = test_request.get('test_id')
                updated_data, status = SOPSRequest().proxy_port_normalisation_request(request_response_object, test_data=json)
                if status.valid:
                    proxy_port_details = updated_data.get('proxy_port_details')
                    proxy_setup_value = proxy_port_details.get('proxy_setup_value')
                    proxy_dict[proxy_name][proxy_setup_key] = proxy_setup_value


    





from scrapy.utils.response import response_httprepr
from scrapy.http import Response
from random import randint

class ResponseValidator(object):

    def __init__(self):
        pass
    
    @staticmethod
    def validate(request_response_object, response, domain_tests=None, generic_tests=None, geotargeting_tests=None):
        if domain_tests is not None:
            for test in domain_tests:
                if ResponseValidator.run_validation_test(response, test.get('validation_tests', [])) is False:
                    request_response_object.failed_validation_test(test)
                    break
        
        if generic_tests is not None:
            for test in generic_tests:
                if ResponseValidator.run_validation_test(response, test.get('validation_tests', [])) is False:
                    request_response_object.failed_validation_test(test)
                    break


    @staticmethod
    def run_validation_test(response, test_array):
        """
        Returns True if test is passed, False if test is failed.
        """
        fail_counter = 0
        for test in test_array:

            if test.get('test_type') == 'bytes_check':
                if ResponseValidator.bytes_check(response, test.get('threshold', 0), test.get('comparison_type')):
                    fail_counter += 1
                else: return True

            if test.get('test_type') == 'response_length_check':
                if ResponseValidator.response_length_check(response, test.get('threshold', 0), test.get('comparison_type')):
                    fail_counter += 1
                else: return True
            
            if test.get('test_type') == 'string_check' and test.get('test_location') == 'body':
                if ResponseValidator.string_check(response, test.get('text_check', ''), test.get('comparison_type'), text_slice=test.get('text_slice')):
                    fail_counter += 1
                else: return True
            
            if test.get('test_type') == 'string_check' and test.get('test_location') == 'user_agent':
                pass

            if test.get('test_type') == 'string_check' and test.get('test_location') == 'url':
                if ResponseValidator.string_check(response.url, test.get('text_check', ''), test.get('comparison_type'), text_slice=test.get('text_slice')):
                    fail_counter += 1
                else: return True

            if test.get('test_type') == 'xpath_check' and test.get('test_location') == 'body':
                if ResponseValidator.xpath_check(response, test.get('xpath_selector', ''), test.get('text_check', ''), test.get('comparison_type')):
                    fail_counter += 1
                else: return True
            
            if test.get('test_type') == 'css_selector_check' and test.get('test_location') == 'body':
                if ResponseValidator.css_selector_check(response, test.get('css_selector', ''), test.get('text_check', ''), test.get('comparison_type')):
                    fail_counter += 1
                else: return True


        if fail_counter == len(test_array):
            return False
        return True


    @staticmethod
    def get_domain_tests(request_response_object, domains):
        domain_details = domains.get(request_response_object.get_domain()) 
        if domain_details is not None:
            return domain_details.get('validation_details')
        return None


    @staticmethod
    def failed_scan(request_response_object, domains):
        domain_details = domains.get(request_response_object.get_domain()) 
        if domain_details is not None:
            failed_scan_ratio = domain_details.get('failed_generic_scan')
            if failed_scan_ratio == 0: return False
            if failed_scan_ratio == 1: return True
            if randint(1, failed_scan_ratio) == 1: return True
        return False


    @staticmethod
    def string_check(response, text_check, comparison, text_slice=None):
        if text_check == '': return False
        text = response.text if isinstance(response, Response) else response
        if text_slice is not None:
            text = ResponseValidator.string_slice(text, text_slice)
        if comparison == 'contains' and text_check in text:
            return True
        elif comparison == 'not_contain' and text_check not in text:
            return True
        return False

    @staticmethod
    def string_slice(text, text_slice):
        if text_slice.get('active'):
            if text_slice.get('slice_type') == 'first':
                return text[:text_slice.get('slice_upper_threshold', len(text))]
            if text_slice.get('slice_type') == 'last':
                return text[text_slice.get('slice_lower_threshold', 0):]
            if text_slice.get('slice_type') == 'range':
                return text[text_slice.get('slice_lower_threshold', 0):text_slice.get('slice_upper_threshold', len(text))]
        return text


    @staticmethod
    def xpath_check(response, xpath_selector, text_check, comparison):
        if xpath_selector == '' or text_check == '': return False
        if comparison == 'equals':
            if response.xpath(xpath_selector).get() == text_check:
                return True
        elif comparison == 'not_equal':
            if response.xpath(xpath_selector).get() != text_check:
                return True
        elif comparison == 'contains':
            text = response.xpath(xpath_selector).get()
            return ResponseValidator.string_check(text, text_check, comparison)
        elif comparison == 'not_contain':
            text = response.xpath(xpath_selector).get()
            return ResponseValidator.string_check(text, text_check, comparison)
        return False


    @staticmethod
    def css_selector_check(response, css_selector, text_check, comparison):
        if comparison == 'equals':
            if response.css(css_selector).get() == text_check:
                return True
        elif comparison == 'not_equal':
            if response.css(css_selector).get() != text_check:
                return True
        elif comparison == 'contains':
            text = response.css(css_selector).get()
            return ResponseValidator.string_check(text, text_check, comparison)
        elif comparison == 'not_contain':
            text = response.css(css_selector).get()
            return ResponseValidator.string_check(text, text_check, comparison)
        return False
    

    @staticmethod
    def bytes_check(response, threshold, comparison):
        if threshold == 0: return False
        reslen = len(response_httprepr(response))
        return ResponseValidator.comparison_operators(reslen, threshold, comparison)


    @staticmethod
    def response_length_check(response, threshold, comparison):
        if threshold == 0: return False
        response_text_length = len(response.text)
        return ResponseValidator.comparison_operators(response_text_length, threshold, comparison)
    

    @staticmethod
    def comparison_operators(value, threshold, comparison):
        if comparison == 'less_than':
            return value < threshold
        if comparison == 'less_than_equal':
            return value <= threshold
        if comparison == 'greater_than':
            return value > threshold
        if comparison == 'greater_than_equal':
            return value >= threshold
        if comparison == 'equals':
            return value == threshold
        if comparison == 'not_equal':
            return value != threshold
        return False


   


    







         
    



from scrapy import signals

from scrapeops_scrapy.core.core import ScrapeopsCore 
from scrapeops_scrapy.signals import scrapeops_signals


class ScrapeOpsMonitor(ScrapeopsCore):

    def __init__(self, crawler):
        ScrapeopsCore.__init__(self)
        self.crawler = crawler

    @classmethod
    def from_crawler(cls, crawler):
        ext = cls(crawler)

        # connect the extension object to signals
        crawler.signals.connect(ext.spider_opened,
                                signal=signals.spider_opened)
        crawler.signals.connect(ext.spider_closed,
                                signal=signals.spider_closed)

        crawler.signals.connect(ext.log_request,
                                signal=signals.request_reached_downloader)

        crawler.signals.connect(ext.log_response,
                                signal=signals.response_downloaded)
        

        crawler.signals.connect(ext.log_response_middleware, 
                                signal=scrapeops_signals.scrapeops_response_recieved)

        crawler.signals.connect(ext.log_exception, 
                                signal=scrapeops_signals.scrapeops_exception_recieved)

        crawler.signals.connect(ext.item_scraped, 
                                signal=signals.item_scraped)

        crawler.signals.connect(ext.item_dropped, 
                                signal=signals.item_dropped)

        crawler.signals.connect(ext.item_error, 
                                signal=signals.item_error)

        crawler.signals.connect(ext.response_rejected, 
                                signal=scrapeops_signals.scrapeops_response_rejected)

        return ext

    def spider_opened(self, spider):
        self.start_sdk(spider=spider, crawler=self.crawler)

    def spider_closed(self, spider, reason):
        self.close_sdk(spider=spider, reason=reason)

    def log_request(self, request, spider):
        self.request_stats(request=request)

    def log_response(self, response, request, spider):
        if self.scrapeops_middleware_enabled() is False:
            self.response_stats(request=request, response=response)

    def log_response_middleware(self, request=None, response=None, spider=None):
        if self.scrapeops_middleware_enabled():
            self.response_stats(request=request, response=response)

    def log_exception(self, request=None, spider=None, exception_class=None):
        if self.scrapeops_middleware_enabled():
            self.exception_stats(request=request, exception_class=exception_class)
        
    def item_scraped(self, item, response, spider): 
        self.item_stats(signal_type='item_scraped', item=item, response=response, spider=spider) 

    def item_dropped(self, item, response, spider): 
        self.item_stats(signal_type='item_dropped', item=item, response=response, spider=spider) 
    
    def item_error(self, item, response, spider): 
        self.item_stats(signal_type='item_error', item=item, response=response, spider=spider) 

    def response_rejected(self, spider=None, response=None, reason=None): 
        pass



from scrapeops_scrapy.core.api import SOPSRequest 
from scrapeops_scrapy.utils import utils

class ErrorLogger(object):

    ERROR_LOGGER_ACTIVE = True

    def __init__(self, spider, crawler, spider_settings, server_hostname, server_ip, start_time, log_file):
        self.spider = spider
        self.crawler = crawler
        self.bot_name = crawler.settings.get('BOT_NAME', 'None')
        self.spider_settings = spider_settings
        self.server_hostname = server_hostname
        self.server_ip = server_ip
        self.start_time = start_time
        self.log_file = log_file
        self._error_history = []
        self.job_group_name = None
        self.job_group_id = None

    def update_error_logger(self, job_name, job_id):
        self.job_group_name = job_name
        self.job_group_id = job_id

    def log_error(self, reason=None, error=None, data=None, request_type=None):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            self._error_history.append({
                'time': utils.current_time(),
                'reason': reason,
                'error': str(error),
                'data': data,
                'request_type': request_type,
            })
    

    def send_error_report(self, error_type=None, body=None, log_data=False):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            try:
                data, status = SOPSRequest().error_report_request(error_type=error_type, body=body)
                if status.valid:
                    if log_data and self.log_file is not None and data.get('sdk_error_id') is not None:
                        with open(self.log_file, 'rb') as f:
                            post_body = {
                                            'sops_sdk': 'scrapy',
                                            'spider_name': self.spider.name,
                                            'job_group_id': self.job_group_id,
                                            'job_group_name': self.job_group_name,
                                            'sdk_error_id': data.get('sdk_error_id')
                                        }
                            _, status = SOPSRequest().error_report_request(error_type=error_type, body=post_body, files={'file': f})  
                            if status.valid is False:
                                self.log_error(reason='send_error_logs_failed', error=status.error)  

                if status.valid is False:
                        self.log_error(reason='send_error_report_failed', error=status.error) 
            except Exception:
                pass


    def sdk_error_close(self, reason=None, error=None, request_type=None, data=None):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            self.log_error(reason=reason, error=error, data=data, request_type=request_type)
            error_data = {
                'final_reason': reason,
                'sops_sdk': 'scrapy',
                'spider_name': self.spider.name,
                'bot_name': self.bot_name, 
                'server_ip': self.server_ip,
                'server_hostname': self.server_hostname,
                'job_group_id': self.job_group_id,
                'job_group_name': self.job_group_name,
                'job_args': utils.get_args(),
                'job_start_time': self.start_time,
                'sops_scrapeops_version': utils.get_scrapeops_version(),
                'sops_scrapy_version': utils.get_scrapy_version(),
                'sops_python_version': utils.get_python_version(),
                'sops_system_version': utils.get_system_version(),
                'sops_middleware_enabled': utils.scrapeops_middleware_installed(self.spider_settings),
                'error_history': self._error_history,
            }
            
            self.send_error_report(error_type='sdk_close', body=error_data, log_data=True)
    

         





from scrapy.utils.python import global_object_name

from scrapeops_scrapy.signals import scrapeops_signals


class ScrapeOpsStats:

    def __init__(self):
        pass

    def process_response(self, request, response, spider):
        spider.crawler.signals.send_catch_log(
            signal=scrapeops_signals.scrapeops_response_recieved, 
            request=request, 
            response=response, 
            spider=spider)
        return response

    def process_exception(self, request, exception, spider):
        ex_class = global_object_name(exception.__class__)
        spider.crawler.signals.send_catch_log(
            signal=scrapeops_signals.scrapeops_exception_recieved, 
            request=request, 
            spider=spider,
            exception_class=ex_class)


from scrapeops_scrapy.utils import utils
from scrapeops_scrapy.core.error_logger import ErrorLogger
from scrapeops_scrapy.core.api import SOPSRequest
from scrapeops_scrapy.normalizer.middleware import RequestResponseMiddleware 
from scrapeops_scrapy.validators.item_validator import ItemValidator 
from scrapeops_scrapy.stats.failed_urls import FailedUrlsHandler
from scrapeops_scrapy.core.model import SDKData



class SDKSetup(SDKData):

    def __init__(self):
        SDKData.__init__(self)


    def initialize_SDK(self, spider, crawler=None):

        ## Spider Data
        self.spider = spider
        self.crawler = crawler
        self.spider_name = spider.name
        self.project_name = crawler.settings.get('PROJECT', None) 
        self.bot_name = crawler.settings.get('BOT_NAME', None)
        self.retry_enabled = crawler.settings.get('RETRY_ENABLED', None) 
        self.retry_times = crawler.settings.get('RETRY_TIMES', None) 
        self.log_file = crawler.settings.get('LOG_FILE', None) 
        self.allowed_response_codes = crawler.settings.get('HTTPERROR_ALLOWED_CODES', []) 
        self.check_spider_attributes(spider)
        self.get_settings(spider)

        ## Job Data
        self.job_args = utils.get_args()
        self.job_group_name = crawler.settings.get('SCRAPEOPS_JOB_NAME', self.get_job_name())
        self.job_group_uuid = crawler.settings.get('SCRAPEOPS_JOB_GROUP_IDENTIFIER', self.get_uuid()) ## Multi-server
        self.job_group_version = crawler.settings.get('SCRAPEOPS_JOB_VERSION', self.get_job_version())
        self.check_scrapeops_triggered_job(crawler)

        ## System Settings
        self._scrapeops_sdk_version = utils.get_scrapeops_version()
        self._scrapeops_scrapy_version = utils.get_scrapy_version()
        self._scrapeops_python_version = utils.get_python_version()
        self._scrapeops_system_version = utils.get_system_version()
        self.get_server_details()

        ## SDK Setup Data
        self._scrapeops_middleware = utils.scrapeops_middleware_installed(self.spider_settings)
        self._scrapeops_job_start = crawler.settings.get('SCRAPEOPS_JOB_START', utils.current_time()) ## Multi-server
        self._scrapeops_server_id = crawler.settings.get('SCRAPEOPS_SERVER_ID', "-1")
        self._scrapeops_debug_mode = crawler.settings.get('SCRAPEOPS_DEBUG_MODE', False)
        self._scrapeops_settings_exclusion_list = crawler.settings.get('SCRAPEOPS_SETTINGS_EXCLUSION_LIST', [])
        self._scrapeops_export_scrapy_logs = self.get_export_logs(crawler) 

        ## SOPS API
        SOPSRequest.SCRAPEOPS_ENDPOINT = crawler.settings.get('SCRAPEOPS_ENDPOINT', 'https://api.scrapeops.io/')
        SOPSRequest.API_KEY = self._scrapeops_api_key = crawler.settings.get('SCRAPEOPS_API_KEY', None)
        SOPSRequest.SCRAPEOPS_LOGGING_DATA = {'logging_data': self.logging_data()}
        
        ## Middlewares
        self.initialize_middlewares()
        self.initialize_error_logger()

    

    def initialize_middlewares(self):
        if self.item_validation_middleware is None: 
            self.item_validation_middleware = ItemValidator()

        if self.failed_url_middleware is None:
            self.failed_url_middleware = FailedUrlsHandler()
    

    def initialize_error_logger(self):
        self._error_logger = ErrorLogger(
                                        self.spider, 
                                        self.crawler,
                                        self.spider_settings, 
                                        self.server_hostname, 
                                        self.server_ip,
                                        self.start_time,
                                        self.log_file)
        

    def initialize_job_details(self, data):
        self.job_id = data.get('job_id')
        self.job_group_name = data.get('job_group_name', self.job_group_name)
        self.job_group_id = SOPSRequest.JOB_GROUP_ID = data.get('job_group_id')
        self.spider_id= data.get('spider_id')
        self.server_id= data.get('server_id')
        self.project_id= data.get('project_id')
        self.multi_server = data.get('multi_server', False)
        self._period_frequency = data.get('stats_period_frequency')
        self._period_freq_list = data.get('stats_period_freq_list')
        self._error_logger.update_error_logger(self.job_group_name, self.job_group_id)
        self.update_sdk_settings(data)
        self.initialize_normalizer_middleware(data)
        SOPSRequest.SCRAPEOPS_LOGGING_DATA = {'logging_data': self.logging_data()}
        

    def initialize_normalizer_middleware(self, data=None):
        if data is not None:
            self._proxy_apis = data.get('proxy_apis', {})
            self._generic_validators = data.get('generic_validators', [])
        if self.request_response_middleware is None:
            self.request_response_middleware = RequestResponseMiddleware(self.job_group_id, 
                                                                            self._proxy_apis, 
                                                                            self._generic_validators, 
                                                                            self._error_logger,
                                                                            self.allowed_response_codes)


    def update_sdk_settings(self, data):
        self._sdk_active = data.get('sdk_active', self._sdk_active) 
        self.multi_server = data.get('multi_server', self.multi_server)
        self.job_group_name = data.get('job_group_name', self.job_group_name)
        self._scrapeops_export_scrapy_logs = data.get('scrapeops_export_scrapy_logs', self._scrapeops_export_scrapy_logs) 

        ## SOPS API Endpoints
        SOPSRequest.SCRAPEOPS_ENDPOINT = data.get('scrapeops_endpoint', SOPSRequest.SCRAPEOPS_ENDPOINT) 
        SOPSRequest.SCRAPEOPS_API_VERSION = data.get('scrapeops_api_version', SOPSRequest.SCRAPEOPS_API_VERSION) 

        ## Normalisation Middleware
        RequestResponseMiddleware.PROXY_DOMAIN_NORMALIZATION = data.get('proxy_domain_normalization', RequestResponseMiddleware.PROXY_DOMAIN_NORMALIZATION) 
        RequestResponseMiddleware.PROXY_ALERTS = data.get('proxy_alerts', RequestResponseMiddleware.PROXY_ALERTS)
        RequestResponseMiddleware.RESPONSE_VALIDATION = data.get('response_validation', RequestResponseMiddleware.RESPONSE_VALIDATION)

        ## Item Validation Middleware
        ItemValidator.ITEM_COVERAGE_ENABLED = data.get('item_coverage_enabled', ItemValidator.ITEM_COVERAGE_ENABLED) 
        ItemValidator.INVALID_ITEM_URLS_LOGGING_ENABLED = data.get('ivalid_item_coverage_url_logging_enabled', ItemValidator.INVALID_ITEM_URLS_LOGGING_ENABLED) 
        ItemValidator.MAX_ITEM_URLS = data.get('max_item_urls', ItemValidator.MAX_ITEM_URLS) 

        ## Failed URL Middleware
        FailedUrlsHandler.FAILED_URL_LOGGER_ENABLED = data.get('FAILED_URL_LOGGER_ENABLED', FailedUrlsHandler.FAILED_URL_LOGGER_ENABLED) 
        FailedUrlsHandler.LOG_MISSED_URLS = data.get('log_missed_urls', FailedUrlsHandler.LOG_MISSED_URLS) 
        FailedUrlsHandler.MAX_LOGGED_URLS = data.get('max_failed_urls', FailedUrlsHandler.MAX_LOGGED_URLS) 

        ## Error Logger
        ErrorLogger.ERROR_LOGGER_ACTIVE = data.get('error_logger', ErrorLogger.ERROR_LOGGER_ACTIVE)


    

    


    







    

    
         
    




class ScrapeOpsMissingAPIKey(Exception):
    """Indicates that no ScrapeOps API key added"""
    def __init__(self):
        self.message = 'No ScrapeOps API key defined.'
        super().__init__(self.message)
    
    def __str__(self):
        return f'ScrapeOpsMissingAPIKey: {self.message}'


class ScrapeOpsAPIResponseError(Exception):
    
    def __init__(self):
        super().__init__()


class DecodeError(Exception):
    pass


import requests

class ScrapeOpsTest:

    def __init__(self):
        self.active = True
        self.test_id = None
    
    def test_active(self):
        if self.active is True:
            return True
        return False

    def get_test_id(self):
        return self.test_id 

    def generate_test_id(self):
        response = requests.post('https://api.scrapeops.io/api/v1/start_test?api_key=1234&sdk_type=scrapy')
        data = response.json() 
        self.test_id = data.get('test_id', None)
        return self.test_id

    @staticmethod
    def generate_test_settings():
        return {
            'RETRY_TIMES': 0,
            'RETRY_ENABLED': False,
        }

    


from setuptools import setup, find_packages


VERSION = '0.3.0'
DESCRIPTION = 'Scrapeops Scrapy SDK, is a monitoring tool for your Scrapy spiders.'

setup(name='scrapeops_scrapy',
      description="This is an extension for Scrapy spiders to enable scraping monitoring, statistics, alerting, scheduling and data validation.",
      long_description="This an extension for your Scrapy spiders that gives you all the scraping monitoring, statistics, alerting, scheduling and data validation you will need straight out of the box.",
      author="ScrapeOps",
      author_email="info@scrapeops.io",
      version="0.3.0",
      license="BSD",
      url="https://github.com/ScrapeOps/scrapeops-scrapy-sdk",
      packages=find_packages(),
      install_requires=[
          "tld==0.12.4",
          "requests==2.24.0",
          "json5==0.9.5",
          "urllib3==1.25.10",
          ],
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
      ],
      python_requires=">=3.6",
      )
